package at.ac.uibk.gitsearch.service;

import static org.assertj.core.api.Assertions.assertThat;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.doNothing;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.when;

import at.ac.uibk.gitsearch.GitsearchApp;
import at.ac.uibk.gitsearch.domain.User;
import at.ac.uibk.gitsearch.domain.UserWatchList;
import at.ac.uibk.gitsearch.domain.WatchListEntry;
import at.ac.uibk.gitsearch.domain.enumeration.CheckFrequency;
import at.ac.uibk.gitsearch.repository.UserRepository;
import at.ac.uibk.gitsearch.repository.UserWatchListRepository;
import at.ac.uibk.gitsearch.repository.search.MetaDataRepository;
import at.ac.uibk.gitsearch.repository.search.testESService.ElasticSearchTestConfiguration;
import at.ac.uibk.gitsearch.service.dto.AdminUserDTO;
import com.fasterxml.jackson.core.JsonGenerationException;
import com.fasterxml.jackson.databind.JsonMappingException;
import java.io.IOException;
import java.time.Instant;
import java.time.LocalDateTime;
import java.util.HashSet;
import java.util.List;
import java.util.Optional;
import java.util.concurrent.TimeUnit;
import javax.mail.internet.MimeMessage;
import org.codeability.sharing.plugins.api.search.SearchResultDTO;
import org.elasticsearch.node.NodeValidationException;
import org.junit.jupiter.api.AfterAll;
import org.junit.jupiter.api.AfterEach;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.Timeout;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.Mock;
import org.mockito.MockitoAnnotations;
import org.mockito.Spy;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.context.MessageSource;
import org.springframework.data.auditing.DateTimeProvider;
import org.springframework.mail.javamail.JavaMailSenderImpl;
import org.springframework.transaction.annotation.Transactional;
import org.thymeleaf.spring5.SpringTemplateEngine;
import tech.jhipster.config.JHipsterProperties;

/**
 * Integration tests for {@link UserWatchListService}.
 */
@SpringBootTest(classes = GitsearchApp.class)
@Transactional
public class InfoMailServiceIT {

    private static final String DEFAULT_LOGIN = "TestUser1";

    private static final String DEFAULT_EMAIL = "TestUser1@localhost";

    private static final String DEFAULT_FIRSTNAME = "Test";

    private static final String DEFAULT_LASTNAME = "User1";

    private static final String DEFAULT_IMAGEURL = "http://placehold.it/50x50";

    private static final String DEFAULT_LANGKEY = "en";

    private static final String TEST_EXERCISE_ID = "[1]";

    @Autowired
    private JHipsterProperties jHipsterProperties;

    @Autowired
    private MessageSource messageSource;

    @Autowired
    private SpringTemplateEngine templateEngine;

    @Autowired
    private UserRepository userRepository;

    @Autowired
    private UserService userService;

    @Autowired
    private UserWatchListService userWatchListService;

    @Autowired
    private UserWatchListRepository userWatchListRepository;

    @Autowired
    private MetaDataRepository metaDataRepository;

    @Spy
    private JavaMailSenderImpl javaMailSender;

    @Captor
    private ArgumentCaptor<MimeMessage> messageCaptor;

    private MailService mailService;

    @Mock
    private DateTimeProvider dateTimeProvider;

    private User user;

    @Autowired
    ElasticSearchTestConfiguration elasticSearchTestConfiguration;

    static ElasticSearchTestConfiguration staticElasticSearchTestConfiguration;

    @BeforeEach // must be started as BeforeEach, in order to autowire the ElasticSearchTestConfiguration
    @Timeout(value = 2, unit = TimeUnit.MINUTES)
    public void setUpESServer() throws IOException, NodeValidationException {
        if (staticElasticSearchTestConfiguration == null) {
            elasticSearchTestConfiguration.startTestNode();
            staticElasticSearchTestConfiguration = elasticSearchTestConfiguration;
        }
    }

    @AfterAll
    public static void shutDownESServer() {
        if (staticElasticSearchTestConfiguration != null) {
            staticElasticSearchTestConfiguration.stopTestContainer();
            staticElasticSearchTestConfiguration = null;
        }
    }

    @BeforeEach
    public void init() {
        AdminUserDTO userDTO = new AdminUserDTO();
        userDTO.setLogin(DEFAULT_LOGIN);

        userDTO.setActivated(true);
        userDTO.setEmail(DEFAULT_EMAIL);
        userDTO.setFirstName(DEFAULT_FIRSTNAME);
        userDTO.setLastName(DEFAULT_LASTNAME);
        userDTO.setImageUrl(DEFAULT_IMAGEURL);
        userDTO.setLangKey(DEFAULT_LANGKEY);
        userDTO.setAuthorities(new HashSet<>());
        userDTO.getAuthorities().add("sharing");

        when(dateTimeProvider.getNow()).thenReturn(Optional.of(LocalDateTime.now()));

        userDTO.setActivated(false);
        user = userService.createUser(userDTO);

        UserWatchList wlDAILY = new UserWatchList("someTestWatchList", CheckFrequency.DAILY, user);

        WatchListEntry wle1 = new WatchListEntry();
        wle1.setExerciseId(TEST_EXERCISE_ID);
        wle1.setExerciseName("Unused in this test");
        wlDAILY.getWatchListEntries().add(wle1);

        userWatchListRepository.save(wlDAILY);

        UserWatchList wlMONTHLY = new UserWatchList("someTestWatchList", CheckFrequency.MONTHLY, user);

        WatchListEntry wle2 = new WatchListEntry();
        wle2.setExerciseId(TEST_EXERCISE_ID);
        wle2.setExerciseName("Unused in this test");
        wlMONTHLY.getWatchListEntries().add(wle2);

        userWatchListRepository.save(wlMONTHLY);
    }

    @BeforeEach
    public void setup() {
        MockitoAnnotations.openMocks(this);
        doNothing().when(javaMailSender).send(any(MimeMessage.class));
        mailService =
            new MailService(jHipsterProperties, javaMailSender, messageSource, templateEngine, userRepository, userWatchListService);
    }

    private String original = null;

    @BeforeEach
    public void setUpChangedExercise() throws JsonGenerationException, JsonMappingException, IOException {
        final SearchResultDTO testExercise = metaDataRepository.getExerciseById(TEST_EXERCISE_ID, Optional.empty());

        testExercise.getProject().setLast_activity_at(Instant.now());

        original = elasticSearchTestConfiguration.updateMetadata(testExercise, "[3]");
    }

    @Test
    @Transactional
    public void testSendInfoMail() throws Exception {
        mailService.sendInfoMails();
        verify(javaMailSender).send(messageCaptor.capture());
        List<MimeMessage> messages = messageCaptor.getAllValues();
        assertThat(messages.size()).isEqualTo(1);

        MimeMessage message = messages.get(0);
        assertThat(message.getAllRecipients()[0].toString()).isEqualTo(user.getEmail());
        assertThat(message.getFrom()[0].toString()).isEqualTo(jHipsterProperties.getMail().getFrom());
        assertThat(message.getContent().toString()).isNotEmpty();
        assertThat(message.getDataHandler().getContentType()).isEqualTo("text/html;charset=UTF-8");
    }

    @AfterEach
    public void resetChangedExercise() throws JsonGenerationException, JsonMappingException, IOException {
        elasticSearchTestConfiguration.resetRawMetadata(original, "[3]");
    }
}
